// public constants
var /* const */ QU = '\''; // Quote
var /* const */ NL = '\n'; // Newline

// private constants
var /* const */ _ET = String.fromCharCode(38); // ampersand
var /* const */ _LT = String.fromCharCode(60); // less than
var /* const */ _GT = String.fromCharCode(62); // greater than

// public variables
var input;
var separator;
var inputs;
var nextInputIndex;
var output;

// private variables
var _title = ''; // to be appended to title
var _input;
var _separator;
var _program;
var _run_error;
var _error_message;
var _output;
var _action;
var _machine_page = 'machine.html';
var _machine_id = 'machine';
var _clone_number = 0; // number of clones made from this page

// public functions

function write() {
  for (var i = 0; i != arguments.length; ++i) {
    output += arguments[i];
  }
}

function writeln() {
  for (var i = 0; i != arguments.length; ++i) {
    output += arguments[i];
  }
  output += NL;
}

function writequ() {
  write(QU);
}

function obj2Str(obj) {
  // Return string with all properties of obj and their values
  // Optional second argument is object with options:
  //   name: string to prepend with ' = '
  //   indent: string used for indentation (default = '')
  //   indentStep: string used to step indentation (default = '  ')
  //   depth: to restrict object expansion depth (default = 0), <0 unrestricted
  //   types: boolean to include type information for multiline (default off)
  //   values: boolean to include values (default on)
  //   multiline: boolean to put properties on separate lines (default on)
  //   functions: boolean to include function bodies (default off)
  var options = { };
  if ( arguments.length >= 2 ) {
    options = arguments[1];
  }
  if ( ! options.indent ) options.indent = '';
  var prefix = options.indent + ( options.name ? options.name + ' = ' : '' );
  if ( ! options.indentStep ) options.indentStep = '  ';
  if ( ! options.depth ) options.depth = 0;
  if ( options.types == undefined ) options.types = false;
  if ( options.values == undefined ) options.values = true;
  if ( options.multiline == undefined ) options.multiline = true;
  if ( options.functions == undefined ) options.functions = false;
  
  function obj2StrRec(obj, indent, depth) {
    var result = '';
    var indentPlus = indent + options.indentStep;
    result += '{';
    var separator = options.multiline ? '\n' + indentPlus : ' ';
    var item, type, value;
    for (var p in obj) {
      value = obj[p];
      type = typeof(value);
      // fix the type (JavaScript's typeof does not recognize array and null)
      if ( type == 'object' ) {
        if ( value ) {
          if ( value instanceof Array ) {
            type = 'array';
          }
        }
        else {
          type = 'null';
        }
      }
      item = String(p);
      if ( options.types && options.multiline ) {
        item +=  ': ' + type;
      }
      if ( options.values ) {
        item += options.multiline ? ' = ' : ': ';
        if ( type == 'array' ) {
          item += '[' + value + ']';
        }
        else if ( (type == 'object') && (depth != 0) ) {
          item += obj2StrRec(value, indentPlus, depth-1);
        }
        else if ( (type != 'function') || options.functions ) {
          item += value;
        }
        else { // (type == 'function') && (! options.functions)
          item += type + ' ...';
        }
      }
      result += separator + item;
      if ( ! options.multiline ) {
        separator = ', ';
      }
    }
    result += options.multiline ? '\n' + indent + '}' : ' }';
    return result;
  }
  
  return prefix + obj2StrRec(obj, options.indent, options.depth);
}

function writeObject(obj) {
  // Write object obj
  // Optional second argument is used as name
  if ( arguments.length >= 1 ) {
    write(obj2Str(obj, { name: arguments[1] }));
  }
  else {
    write(obj2Str(obj));
  }
}

function moreInputs() {
  return nextInputIndex < inputs.length;
}

function readStr() {
  return inputs[nextInputIndex++];  // undefined beyond end
}

function readNum() {
  return Number(inputs[nextInputIndex++]);  // undefined beyond end
}

function isQuote(c) {
  return c == QU;
}

function isNewline(c) {
  return c == NL;
}

// private functions

function _parseURI() {
  // Parse the query parameters of the document's URI.
  // requires: URI is well-formed, in particular, that it consists of
  //   a semicolon separated list of key-value pairs
  // ensures: each key has been assigned the corresponding value
  //   this is particularly useful for keys that are already declared as
  //   global variables, such as, _input, _program, _action, which are
  //   used by _init()
  // Also see: _machine_link to generate such URIs
  var uri = document.location.href; // the URI of the current web page
  var re = /^(([^:/?#]+):)?(\/\/([^/?#]*))?([^?#]*)(\?([^#]*))?(#(.*))?/;
    // re = regular expression to parse URIs
    // inspired by <http://labs.apache.org/webarch/uri/rfc/rfc3986.html#regexp>
  var uri_parsed = uri.match(re); // parse uri
  // scheme    = uri_parsed[2]
  // authority = uri_parsed[4]
  // path      = uri_parsed[5]
  // query     = uri_parsed[7]
  // fragment  = uri_parsed[9]
  var query = uri_parsed[7];
  if ( ! query ) { // no query part
    return;
  }
  var key_value_pairs = query.split(';');
  var kv_re = /^([A-Za-z_]\w*)=(.*)/; // reg expr to parse key-value pair
  for (var i = 0; i != key_value_pairs.length; ++i) {
    var key_value = key_value_pairs[i].match(kv_re);
    if ( key_value.length == 3 ) {
      var key = key_value[1];
      var value = decodeURIComponent(key_value[2]);
      eval ( key + ' = value;' ); // assign value to variable for key
    }
  }
}

function _machine_uri(query) {
  // Return a URI to the machine with given query, being an object with
  // string-valued properties, and each property-value pair
  // is encoded in the URI (cf. _parseURI).
  var uri = _machine_page;
  var sep = '?';
  for (var prop in query) {
    uri += sep + prop + '=' + encodeURIComponent(query[prop]);
    sep = ';';
  }
  uri += '#' + _machine_id;
  return uri;
}

function _machine_link(anchorText, program) {
  // Inject a hyperlink to the machine with given anchor text,
  // instructing it to load the given program.
  // When a 3rd argument is given, this must be an object with
  // string-valued properties, and each property-value pair
  // is encoded in the URI (cf. _parseURI).
  var uri = _machine_page + '?_program=';
  uri += encodeURIComponent(program);
  if (arguments.length >= 3) {
    for (var prop in arguments[2]) {
      uri += ';' + prop + '=' + encodeURIComponent(arguments[2][prop]);
    }
  }
  uri += '#' + _machine_id;
  document.write(anchorText.link(uri));
}

function _tag(s) {
  // Return string with tag for element s, optionally with attributes.
  // requires: s is a string; optional further string arguments
  // ensures: result == tagged s, optionally including other arguments,
  //   separated by spaces
  var result = _LT + s;
  var i = 1;
  while (i != arguments.length) {
    result += ' ' + arguments[i];
    i = i + 1;
  }
  result += _GT;
  return result;
}

function _tt(s, c) {
  return _tag('tt', 'class="'+c+'"') + s + _tag('/tt');
}

function _inject(s) {
  // Inject string s in document inside pre-tags.
  // if a second argument t is given, then add ' class="t"'
  if ( arguments.length == 2 ) {
    document.write(_tag('pre', 'class="' + arguments[1] + '"'));
  }
  else {
    document.write(_tag('pre'));
  }
  document.write("\n");
  for (var i = 0; i != s.length; ++i) {
    var c = s.charAt(i);
    if ( c == _LT ) {
      document.write(_ET, 'lt;');
    }
    else if ( c == _GT ) {
      document.write(_ET, 'gt;');
    }
    else if ( c == _ET ) {
      document.write(_ET, 'amp;');
    }
    else {
      document.write(c);
    }
  }
  document.writeln(_tag('/pre'));
}

function _embedded(s, indent) {
  // Returns a string expression in a string such that eval(_embedded(s)) == s,
  // and s contains no occurences of the characters _LT, _GT, _ET.
  // requires: s is a string, possibly containing newlines,
  //   single and double quotes, and backslashes,
  //   but no other control characters (such as tabs)
  // ensures: that result is a "-delimited multi-line string expression
  //   with value s, indented by 2 spaces (default) or using the optional
  //   second string argument for indenting
  // This function can be used in the machine as writeln(_embed(input))
  // to convert the input into a string S that can be embedded in an
  // HTML page e.g. via var _prog = S; _inject(_prog);
  // Also see: _embed()
  var result = indent + '"';
  var i = 0;
  while (i != s.length) {
    var c = s.charAt(i);
    switch (c) {
      case '"':  result += '\\"'; break;
      case '\n': {
        result += '\\n';
        if ( i + 1 < s.length ) {
          result += '"+' + NL + indent + '"';
        }
        break;
      }
      case '\\': result += '\\\\'; break;
      case _LT: result += '\\074'; break;
      case _GT: result += '\\076'; break;
      case _ET: result += '\\046'; break;
      default: result += c;
    }
    i = i + 1;
  }
  result += '"';
  return result;
}

function _embed() {
  // Output an embedded version of input
  // Optional argument is used to indent
  var indent = '  '; // default
  if ( arguments.length >= 1 ) {
    indent = arguments[0];
  }
  writeln(_embedded(input, indent));
}

function _embed_prog() {
  // Output an embedded version of program
  // Optional argument is used to indent
  var prog = document.machine.program.value;
  var indent = '  '; // default
  if ( arguments.length >= 1 ) {
    indent = arguments[0];
  }
  document.machine.progoutput.value = _embedded(prog, indent);
}

function _append_title(s) {
  document.title += s;
}

function _set_input(s) {
  document.machine.proginput.value = s;
}

function _set_program(s) {
  document.machine.program.value = s;
}

function _set_output(s) {
  document.machine.progoutput.value = s;
}

function _clrinp() {
  _set_input('');
}

function _clrout() {
  _set_output('');
}

function _copy_program_to_input() {
  _set_input(document.machine.program.value);
}

function _copy_output_to_input() {
  _set_input(document.machine.progoutput.value);
}

function _copy_output_to_program() {
  _set_program(document.machine.progoutput.value);
}

function _make_error_message(e, p) {
  // Return error message for exception e in program p
  var result = 'Program execution failed';
  if ( e.sourceURL ) {
    result += ' in ' + e.sourceURL;
  }
  if ( e.line ) {
    result += ' on line ' + e.line;
  }
  result += '.' + NL + NL;
  if ( e.line && ! e.sourceURL ) {
    var lines = _clip_lines(p, e.line, 0);
    if ( lines.substr(-1, 1) != NL ) {
      lines += NL;
    }
    result += lines + NL;
  }
  result += e.name + ': ' + e.message;
  result += NL + NL + 'Click OK for more information.';
  return result;
}

function _show_error() {
  if ( _run_error ) {
    if ( confirm(_error_message) ) {
      alert(obj2Str(_run_error, { name: 'Exception' }));
    }
  }
  else {
    alert('There was no error.');
  }
}

function _output_of(p) {
  // Determine output of program p when executed by the machine.
  // if a second argument is given, then this is used as input
  // if a third argument is given, then this is a separator used to
  //   parse and split the input
  var input_old = input;
  var separator_old = separator;
  var inputs_old = inputs;
  var nextInputIndex_old = nextInputIndex;
  var output_old = output;

  if ( arguments.length >= 2 ) {
    input = arguments[1];
  }
  else {
    input = "";
  }
  if ( arguments.length >= 3 ) {
    separator = eval ( arguments[2] );
    var input_to_split;
    if ( separator == ' ' ) { // treat newlines as space, condense space blocks
      input_to_split = input.replace(/\n/g, ' ');
      input_to_split = input_to_split.replace(/ +/g, ' ');
    }
    else if ( ! isNewline(separator) ) { // ignore newlines
      input_to_split = input.replace(/\n/g, '');
    }
    else {
      input_to_split = input;
    }
    if (input_to_split.length == 0) {
      inputs = [ ];
    }
    else {
      inputs = input_to_split.split(separator);
    }
  }
  else { // no separator given
    inputs = input.split();
  }
  nextInputIndex = 0;
  var result;
  output = '';
  _run_error = undefined;
  try {
    eval(p);
    result = output;
  }
  catch (e) {
    _run_error = e;
    _error_message = _make_error_message(e, p);
    _show_error();
    result = undefined;
  }
  input = input_old;
  separator = separator_old;
  inputs = inputs_old;
  nextInputIndex = nextInputIndex_old;
  output = output_old;
  return result;
}

function _clip_lines(s, n, k) {
  // return line n from s, with up to k lines before/after it
  // line numbering starts at 1
  var result = '';
  var line = 1; // number of current line
  for (var i = 0; i != s.length; ++i) {
    var c = s.charAt(i);
    if ( (n - k <= line) && (line <= n + k) ) {
      result += c;
    }
    if ( c == NL ) {
      ++line;
    }
  }
  return result;
}

function _run() {
  var inp = document.machine.proginput.value;
  var sep = document.machine.separator.value;
  var prog = document.machine.program.value;
  var out = _output_of(prog, inp, sep);
  if ( out ) {
    document.machine.progoutput.value += out;
  }
  document.machine.progoutput.focus();
  return out != undefined;
}

function _challenge() {
  if ( document.machine.program.value == '' ) {
    alert('Program is empty!');
    return;
  }
  _clrinp();
  _clrout();
  if ( _run() ) {
    _cmp();
  }
}

function _diff(s, t) {
  // Determine first difference between s and t.
  // assumes that s and t differ
  var result = '';
  var splits = s.split(NL);
  var splitt = t.split(NL);
  if ( s.length != 0 ) {
    if ( isNewline(s.charAt(s.length - 1)) ) {
      splits.pop();
    }
    else {
      result += 'No newline at end of first argument'+NL;
    }
  }
  if ( t.length != 0 ) {
    if ( isNewline(t.charAt(t.length - 1)) ) {
      splitt.pop();
    }
    else {
      result += 'No newline at end of second argument'+NL;
    }
  }
  var i = 0;
  while ( (i != splits.length) && (i != splitt.length) ) {
    if ( splits[i] != splitt[i] ) {
      result += 'First difference on line '+(i+1)+':'+NL+
        _LT+' '+splits[i]+NL+
        _GT+' '+splitt[i]+NL;
      return result;
    }
    ++i;
  }
  if ( i != splits.length ) {
    result += 'First difference on line '+(i+1)+':'+NL+
        _LT+' '+splits[i]+NL;
    return result;
  }
  if ( i != splitt.length ) {
    result += 'First difference on line '+i+':'+NL+
        _GT+' '+splitt[i]+NL;
    return result;
  }
  if ( result != '' ) {
    return result;
  }
  else {
    return 'EQUAL'; // should not happen
  }
}

function _cmp() {
  // Compare program and its output.
  result = '';
  if ( document.machine.program.value == document.machine.progoutput.value ) {
    result += 'OK, Program = Output';
  }
  else {
    result += 'Oh oh, Program and Output DIFFER!'+NL+
      _diff(document.machine.program.value,
        document.machine.progoutput.value);
  }
  alert(result);
}

function _clone() {
  _clone_number += 1;
  var query = {
    _title: _title + '.' + _clone_number,
    _input: document.machine.proginput.value,
    _separator: document.machine.separator.value,
    _program: document.machine.program.value,
    _output: document.machine.progoutput.value
  };
  window.open(_machine_uri(query));
}

function _init() {
  // Initialize the machine.
  _run_error = undefined;
  _error_message = '';
  _parseURI();
  if ( _title ) {
    _append_title(_title);
  }
  if ( _input ) {
    _set_input(_input);
  }
  if ( _separator ) {
    document.machine.separator.value = _separator;
  }
  if ( _program ) {
    _set_program(_program);
  }
  if ( _output ) {
    _set_output(_output);
  }
  if ( _action ) {
    //alert( _action );
    eval ( _action );
  }
}

